/**
 * Funções para trabalhar com blobs: desenhar a silhueta, texturizar, simplificar os polígonos, etc
 *
 * ATENÇAO: O código assume a existencia de duas variaveis globais: 
 * opencv - o objecto do tipo OpenCV ( assume-se uma captura de 320x200 pixeis).
 * blobs - as blobs detectadas pelo opencv (resultado da invocacao de 'blobs = opencv.blobs();'
 * 
 *
 * Jorge Cardoso (Janeiro 2009)
 * http://processing.jorgecardoso.eu
 */

/** 
 * Desenha a blob na posição original e com o tamanho original.
 */
public void drawBlob(Blob b) {
  drawBlob(b, 0, 0, 320, 240);
}

/** 
 * Desenha a blob na posição original e com o tamanho original.
 */
public void drawBlob() {
  if (blobs.length > 0) {
    drawBlob(blobs[0], 0, 0, 320, 240);
  }
}

/**
 * Desenha a blob na posicao especificada (x, y) e com o tamanho especificado (width, height).
 */
public void drawBlob(float x, float y, float width, float height) {
  if (blobs.length > 0) {
    drawBlob(blobs[0], x, y, width, height);
  }
}


/**
 * Desenha a blob na posicao especificada (x, y), com o tamanho especificado (width, height), com um numero
 * maximo de vertices (maxVertices) e com um tamanho minimo para cada aresta (minSegLength).
 *
 */
public void drawBlob(Blob b, float x, float y, float width, float height) {

  Point[] points = b.points; 

  //rect(blobs[0].rectangle.x, blobs[0].rectangle.y, blobs[0].rectangle.width, blobs[0].rectangle.height);
  beginShape(POLYGON);
  for (int i = 0; i < points.length; i++) {
    if (i > 0 ) { 
      vertex(points[i].x*width/320.0+x, points[i].y*height/240.0+y);
    }
  }
  
  endShape(CLOSE);

}

/** 
 * Desenha a blob na posição original e com o tamanho original.
 * A blob é desenhada com uma textura especificada por uma PImage (textura).
 */
public void drawBlobTexture(Blob b, PImage textura) {
  drawBlobTexture(b, textura, 0, 0, 320, 240);
}

/** 
 * Desenha a blob na posição original e com o tamanho original.
 * A blob é desenhada com uma textura especificada por uma PImage (textura).
 */
public void drawBlobTexture(PImage textura) {
  if (blobs.length > 0) {
    drawBlobTexture(blobs[0], textura, 0, 0, 320, 240);
  }
}


/**
 * Desenha a blob na posicao especificada (x, y) e com o tamanho especificado (width, height).
 * A blob é desenhada com uma textura especificada por uma PImage (textura).
 */
public void drawBlobTexture(PImage textura, float x, float y, float width, float height) {
  if (blobs.length > 0) {
    drawBlobTexture(blobs[0], textura, x, y, width, height);
  }
}

/**
 * Desenha a blob na posicao especificada (x, y), com o tamanho especificado (width, height), com um numero
 * maximo de vertices (maxVertices) e com um tamanho minimo para cada aresta (minSegLength). 
 * A blob é desenhada com uma textura especificada por uma PImage (textura).
 */
public void drawBlobTexture(Blob b, PImage textura, float x, float y, float width, float height) {

  //strokeWeight(3);

  // Point[] points = detail(smooth(blobs[0].points, 10, 10), 100);
  Point[] points = b.points; 
  fill(255);
  noStroke();
  beginShape();
  texture(textura);
  for (int i = 0; i < points.length; i++) {
    if (i > 0 ) { 
      vertex(points[i].x*width/320+x, points[i].y*height/240+y, points[i].x, points[i].y);
    }

  }
  endShape();

}

/**
 * Simplifica os poligonos das blobs detectadas.
 * 
 */
void simplificaBlobs() {
  if (blobs.length > 0 ) {
    for (int i = 0; i < blobs.length; i++) {
      blobs[i].points = detail(blobs[i].points, 150, 10);
    }
  }
}

Point intersecta(Blob b, float x1, float y1, float x2, float y2) {
  return intersecta(b, x1, y1, x2, y2, width, height);
}

/**
 * Intersecta um segmento de recta com a blob.
 * Retorna o ponto de intersecção (mais proximo de x1, y1), ou null se não houver intersecção.
 */
Point intersecta(Blob b, float x1, float y1, float x2, float y2, float width, float height) {
  if (b == null) { 
    return null; 
  }
  x1 = x1*320/width;
  x2 = x2*320/width;
  y1 = y1*240/height;
  y2 = y2*240/height;
  Point2D.Double closest = null;
  float minDistSq = 100000000;
  for (int i = 1; i <= b.points.length; i++) {
    Point2D.Double inter;
    if (i == b.points.length) {
      inter = intersection((float)x1, (float)y1, x2, y2, b.points[i-1].x, b.points[i-1].y, b.points[0].x, b.points[0].y);
    } else {
      inter = intersection((float)x1, (float)y1, x2, y2, b.points[i-1].x, b.points[i-1].y, b.points[i].x, b.points[i].y);
    }
    
    if (inter != null) {      
      if (closest == null) {
        closest = inter;
        minDistSq = distSq(x1, y1, (float)closest.x, (float)closest.y);
      }
      else {
        float distSq = distSq(x1, y1, (float)inter.x, (float)inter.y);
        if (distSq < minDistSq) {
          closest = inter;
          minDistSq = distSq;
        }
      }
    }
  }
  if (closest != null) {
    return new Point((int)(closest.x*width/320), (int)(closest.y*height/240));
  } 
  else {
    return null;
  }
}

/**
 * Intersecta um segmento de recta com a blob principal.
 * Retorna o ponto de intersecção, ou null se não houver intersecção.
 */
Point intersecta(float x1, float y1, float x2, float y2) {
  if (blobs.length > 0 ){
    return intersecta(blobs[0], x1, y1, x2, y2, width, height);
  } 
  else {
    return null;
  }
}


/**
 * Reflecte um ponto (ou melhor, segmento) na blob principal.
 * 
 * O resultado é um MeuVector, com o ponto de reflexão e
 * com a velocidade.
 */
MeuVector reflecte(float xx1, float yy1, float xx2, float yy2) {
  if (blobs.length > 0) {
    if (dentroBlob(xx2, yy2)) {
      float x1 = xx1*320*1.0/width;
      float x2 = xx2*320*1.0/width;
      float y1 = yy1*240*1.0/height;
      float y2 = yy2*240*1.0/height;

      float vx = xx2-xx1;
      float vy = yy2-yy1;

      Point p[] = blobs[0].points;

      int closest = -1;
      float distanceSq = 100000000;
      for (int i=0; i < p.length; i++) {
        float d = distSq(x2, y2, p[i].x, p[i].y);
        if (d < distanceSq) {
          closest = i;
          distanceSq = d;
        }
      }

      Point2D.Double n = normal(p[closest].x, p[closest].y, p[(closest+1)%p.length].x, p[(closest+1)%p.length].y);

      Point2D.Double r = reflection((int)(p[closest].x-vx), (int)(p[closest].y-vy), p[closest].x, p[closest].y, n.x, n.y);


      float magV = mag(vx, vy);
      // println("magV: "  +magV);
      float magR = mag((float)r.x, (float)r.y);
      // println(r + " " + (r.x*r.x+r.y*r.y) + " " + magV);
      MeuVector mv =  new MeuVector((float)(p[closest].x+r.x)*width/320.0, (float)(p[closest].y+r.y)*height/240.0, 
      (float)(magV*r.x/magR), (float)(magV*r.y/magR));
      //println(mv);
      return mv;
    }

  } 
  return null; 
}


/** 
 * Verifica se o ponto x, y está dentro da blob principal
 * Assume-se que o ponto x, y é dado no sistema de coordenadas
 * por omissão (0, 0, width, height).
 */
boolean dentroBlob(float x, float y) {
  return dentroBlob(x, y, width, height);
}

/** 
 * Verifica se o ponto x, y está dentro da blob principal
 * Assume-se que o ponto x, y é dado num sistema de coordenadas
 * com origem em (0,0) e maximos (width, height).
 */
boolean dentroBlob(float x, float y, float width, float height) {
  if (blobs.length > 0) {
    if (insideRect(x*320/width, y*240/height, blobs[0].rectangle.x, 
    blobs[0].rectangle.y, 
    blobs[0].rectangle.width, 
    blobs[0].rectangle.height)) {
      return insidePolygon(x*320/width, y*240/height, blobs[0].points);
    } 
    else {
      return false;
    }

  } 
  else {
    return false;
  } 
}




/**
 * Reduz o detalhe de um polígono.
 * Primeiro aumenta o tamanho das arestas, combinando as arestas existentes (eliminando vertices).
 * Depois, se necessário, descarta vertices para garantir que no final o poligono tem, no maximo,
 * maxVertices.
 */
Point[] detail(Point[] points, int maxVertices, float minSegLength) {
  Point p[] = new Point[points.length];

  float minSegLengthSq = minSegLength*minSegLength;

  float accLength = 0;

  p[0] = points[0];

  int count = 1;
  for (int i = 1; i < points.length; i++) {
    float l = distSq(points[i-1], points[i]);
    accLength += l;
    if (accLength >= minSegLengthSq) {
      p[count] = points[i];
      accLength = 0;
      count++;
    }
  }
  Point p1[]; 
  if (count > maxVertices) {
    p1 = new Point[maxVertices];

    float inc = count*1.0/maxVertices;
    for (int i = 0; i < maxVertices; i++) {
      p1[i] = p[(int)(i*inc)];
      // println(i*inc);
    }
  } 
  else {
    p1 = new Point[count];
    arraycopy(p, p1, count);
  }

  return p1;

}






